/* -------------------------------------------------------------------------
     "chemgen.cc" - Reaction Macro Processor
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2016 by A. Miyoshi, University of Tokyo
                                       created: Oct.  18, 2001
                                        edited: Feb.  18, 2002
                                        edited: Aug.  29, 2003
                                        edited: Dec.  10, 2003
                                        edited: Aug.  21, 2004
                                        edited: Dec.  22, 2008
                                        edited: Feb.  20, 2010
                                        edited: Jan.  26, 2011
                                        edited: Feb.   4, 2011
                                        edited: Jul.   3, 2015
                                   last edited: Mar.   1, 2016
 --------------------------------------------------------------------------
        Brief summary of changes made
         Revision 0.10 -> 0.20:
           1) allow comment lines in 'rxnRList.dat'
           2) enlarge the length of the macros
         Revision 0.20 -> 0.21:
            - extensively modified to ba a part of KUCRS
           1) program name changed from "genReact.c" to "chemgen.cc"
           2) data file names changed:
             rxnRList.dat -> ENVVARNAM/rxn_macr.dat
             th0pre.dat   -> ENVVARNAM/bas_thrm.dat
             $2(rxnout)   -> chem.inp
             $3(thmout)   -> therm.dat
             th1add.dat   -> $2(thin)
           3) disabled #include function
         Revision 0.26 -> 0.27:
           1) Changed not to complain when "bas_rxns.inp" or "bas_thrm.dat"
             does not exist.
         Feb. 4, 2011:
           1) Added --no_append option
         Jul. 3, 2015:
           1) Added --append_nox option
 --------------------------------------------------------------------------
 usage: chemgen rxnin thmin [-l|--no_append|-l--no_append|--append_nox|
         --l--append_nox]
------------------------------------------------------------------------- */

#include "rxnmacro.h"
using namespace std;

#define FL_BASRXN "bas_rxns.inp"        // base mechanism - reactions
#define FL_BASTHM "bas_thrm.dat"        // base mechanism - thermo
#define FL_NOXRXN "nox_rxns.inp"        // NOx mechanism - reactions
#define FL_NOXTHM "nox_thrm.dat"        // NOx mechanism - thermo
#define FN_OUTRXN "_rxn.inp"            // reactions block output
#define FN_OUTSPC "_spc.inp"            // species block output
#define FN_OUTCHM "_chm.inp"            // chem (reactions + species)
#define FN_OUTTHM "_thm.dat"            // thermo block output
#define FX_THERMD "therm.dat"           // extra output = FN_OUTTHM

#define PS_OMIT   10
#define PS_COMNT  20

// --- prototype declarations ---

int parseRxnLine(const string &s, rxnInfo &r, thermTable &t,
 rateMacro &m);
void outElemBlock(ofstream &ofs, set<int> &uAtm, int asLower);
int outSpcThmNew(ofstream &ofs, ofstream &oft, string flnnth,
 thermTable &thmt, int asLower);
int appendFileRxn(ofstream &ofs, string flname, libPath &libp);
int appendFileThm(ofstream &ofs, ofstream &oft, string flname,
 libPath &libp, thermTable &thmt, int asLower);

/* ===== main ========================================================= */
int main(int argc, char *argv[]) {
  static double Ts[NUMTLST] = {300., 400., 500., 600., 800., 1000.,
   1500., 2000., 2500., 3000., 3500., 4000., 5000. };
  int retc, retcCal, atSymLowerCase = false, no_append = false;
  int append_nox = false;
  string clopt, cdate, s, fnBas;
  string flnRxnIn, flnThmIn, flnRxnOut, flnSpcOut, flnChmOut, flnThmOut;
  ifstream ifrxn;
  ofstream ofrx, ofsp, ofth;
  set<int> usedAtoms;
  libPath libp;
  rateMacro ratm;
  thermTable thmt;
  rxnInfo rinf;

  if ((argc < 3) || (argc > 4))
    { cout << "usage: chemgen rxnin thmin "
           << "[-l|--no_append|-l--no_append|--append_nox|--l--append_nox]\n";
     return 1; }
  if (argc == 4) {
    clopt = argv[3];
    if (clopt == "-l") { atSymLowerCase = true; }
    else if (clopt == "--no_append") { no_append = true; }
    else if (clopt == "-l--no_append") { atSymLowerCase = true; no_append = true; }
    else if (clopt == "--append_nox") { append_nox = true; }
    else if (clopt == "-l--append_nox") { atSymLowerCase = true; append_nox = true; }
    else { cout << "usage: chemgen rxnin thmin "
                << "[-l|--no_append|-l--no_append|--append_nox|--l--append_nox]\n";
          return 1; }
  }
  flnRxnIn = argv[1]; flnThmIn = argv[2];
  fnBas = getBasePartEndWith(flnRxnIn, "_chg.inp");
  flnRxnOut = fnBas + FN_OUTRXN; flnSpcOut = fnBas + FN_OUTSPC;
  flnChmOut = fnBas + FN_OUTCHM; flnThmOut = fnBas + FN_OUTTHM;
  thmt.setM(); thmt.appendFrom(FL_BASTHM, libp); thmt.setAllIsUsed();
  if (append_nox) { thmt.appendFrom(FL_NOXTHM, libp); thmt.setAllIsUsed(); }
  thmt.appendFrom(flnThmIn);
  ofrx.open(flnRxnOut.c_str());
  if (!ofrx)
    { cout << "Error writing " << flnRxnOut << ".\n"; return 1; }
  ifrxn.open(flnRxnIn.c_str());
  if (!ifrxn) {
    cout << "Error reading " << flnRxnIn << ".\n"; ofrx.close(); return 1;
  }
  cdate = "(" + strCurDate() + ")";
  ofrx << "reactions      cal/mole  moles\n";
  cout << formatLeft(" reaction", 48) << "      delH      delS\n";
  cout << formatLeft("", 48) << "  kcal/mol cal/K/mol\n";
  while (getOneLine(ifrxn, s) == NORM) {
    if ((retc = parseRxnLine(s, rinf, thmt, ratm)) == ERRE) { break; }
    if (retc == NORM) {
      if ((retcCal = rinf.setRateConst(thmt, ratm, Ts)) == ERRE) { break; }
      rinf.outputToChem(ofrx, Ts);
    } else if (retc == PS_COMNT) { ofrx << s << endl; }
  }
  ifrxn.close();
  if ((retc == ERRE) || (retcCal == ERRE)) { ofrx.close(); return 1; }
  if (!no_append) {
    if (appendFileRxn(ofrx, FL_BASRXN, libp) == ERRE) { return 1; }
    if (append_nox) {
      if (appendFileRxn(ofrx, FL_NOXRXN, libp) == ERRE) { return 1; }
    }
  }
  ofrx << "end\n";
  ofrx.close();
  ofsp.open(flnSpcOut.c_str());
  if (!ofsp)
    { cout << "Error writing " << flnSpcOut << ".\n"; return 1; }
  usedAtoms = thmt.getUsedAtoms();
  outElemBlock(ofsp, usedAtoms, atSymLowerCase);
  ofsp << "species\n";
  ofth.open(flnThmOut.c_str());
  if (!ofth) {
    cout << "Error writing " << flnThmOut << ".\n";
    ofsp.close(); return 1;
  }
  ofth << "thermo all\n" << "   300.000  1000.000  5000.000\n";
  if (outSpcThmNew(ofsp, ofth, flnThmIn, thmt, atSymLowerCase) == ERRE)
    { ofth.close(); ofsp.close(); return 1; }
  if (!no_append) {
    if (appendFileThm(ofsp, ofth, FL_BASTHM, libp, thmt, atSymLowerCase)
        == ERRE) { ofth.close(); ofsp.close(); return 1; }
  }
  if (append_nox) {
    if (appendFileThm(ofsp, ofth, FL_NOXTHM, libp, thmt, atSymLowerCase)
        == ERRE) { ofth.close(); ofsp.close(); return 1; }
  }
  ofth << "end\n";
  ofth.close();
  ofsp << "end\n";
  ofsp.close();
  cout << "Copying " << flnSpcOut << " + " << flnThmOut << " + "
       << flnRxnOut << " -> " << flnChmOut << " - ";
  if (add3Files(flnSpcOut, flnThmOut, flnRxnOut, flnChmOut) == ERRE)
    { cout << "failed.\n"; }
  else { cout << "succeeded.\n"; }
  return 0;
}

/* ----- parse a reaction input line ---------------------------------- */
int parseRxnLine(const string &s, rxnInfo &r, thermTable &t,
 rateMacro &m) {
  static const string spl = " ,\t";
  int ntkns;
  vector<string> vs;

  if ((ntkns = splitToken(s, spl, vs)) == 0) { return PS_OMIT; }
  if (vs[0][0] == '#') { return PS_OMIT; }
  if (vs[0][0] == '!') { return PS_COMNT; }
  if (r.restoreFrom(vs, t, m) == ERRE) { return ERRE; }
  return NORM;
}

/* ----- output elements block ---------------------------------------- */
void outElemBlock(ofstream &ofs, set<int> &uAtm, int asLower) {
  string aSym;
  set<int>::iterator p;
  printHeadingInfo(ofs);
  ofs << "elements\n";
  for (p = uAtm.begin(); p != uAtm.end(); p++) {
    if (asLower) { aSym = strToLower(elmt[*p].atomicSymbol); }
    else { aSym = elmt[*p].atomicSymbol; }
    ofs << ' ' << aSym;
  }
  ofs << endl << "end\n";
}

/* ----- output species and thermo for new species -------------------- */
int outSpcThmNew(ofstream &ofs, ofstream &oft, string flnnth,
 thermTable &thmt, int asLower) {
  int id, skip, slen;
  ifstream ifs;
  string s, nameSpc, sSpOut = "";
  vector<string> vs;

  ifs.open(flnnth.c_str());
  if (!ifs) { cout << " Error reading " << flnnth << ".\n"; return ERRE; }
  while (!ifs.eof()) {
    if (getOneLine(ifs, s) == ERRE) { break; }
    skip = false; slen = s.length();
    if (slen < 3) { skip = true; }
    else {
      if (strToUpper(s.substr(0, 3)) == "END") { break; }
      if (s[0] == '!') { skip = true; }
      else if (slen > 5) {
        if (strToUpper(s.substr(0, 6)) == "THERMO") {
          if (getOneLine(ifs, s) == ERRE) { break; }
          skip = true;
        }
      }
    }
    if (!skip) {
      vs.clear(); vs.push_back(s);
      for (id = 0; id < 3; id++) {
        if (getOneLine(ifs, s) == ERRE) { break; }
        vs.push_back(s);
      }
      nameSpc = strTermBy(vs[0], ' ');
      if (nameSpc.length() > 18) { nameSpc = nameSpc.substr(0, 18); }
      if (thmt[nameSpc].isUsed) {
        sSpOut += (' ' + nameSpc);
        if (sSpOut.length() > 60) { ofs << sSpOut << endl; sSpOut = ""; }
        oft << thmt[nameSpc].fix1stLine(vs[0], asLower) << endl;
        for (id = 1; id < 4; id++)
          { oft << thmt[nameSpc].fixDataLine(vs[id], id + 1) << endl; }
      }
    }
  }
  if (sSpOut.length() > 0) { ofs << sSpOut << endl; }
  ifs.close();
  return NORM;
}

/* ----- append reaction inputs --------------------------------------- */
int appendFileRxn(ofstream &ofs, string flname, libPath &libp) {
  int skip, slen;
  ifstream ifs;
  string s;

  ifs.open(libp.fullPath(flname).c_str());
  if (!ifs) { return NORM; }
  while (!ifs.eof()) {
    if (getOneLine(ifs, s) == ERRE) { break; }
    skip = false; slen = s.length();
    if (slen < 3) { skip = true; }
    else {
      if (strToUpper(s.substr(0, 3)) == "END") { break; }
      if (s[0] == '!') { skip = true; }
      else if (slen > 8) {
        if (strToUpper(s.substr(0, 9)) == "REACTIONS") { skip = true; }
      }
    }
    if (!skip) { ofs << s << endl; }
  }
  ifs.close();
  return NORM;
}

/* ----- append thermo inputs and output to species out --------------- */
int appendFileThm(ofstream &ofs, ofstream &oft, string flname,
 libPath &libp, thermTable &thmt, int asLower) {
  int id, skip, slen;
  ifstream ifs;
  string s, nameSpc, sSpOut = "";
  vector<string> vs;

  ifs.open(libp.fullPath(flname).c_str());
  if (!ifs) { return NORM; }
  while (!ifs.eof()) {
    if (getOneLine(ifs, s) == ERRE) { break; }
    skip = false; slen = s.length();
    if (slen < 3) { skip = true; }
    else {
      if (strToUpper(s.substr(0, 3)) == "END") { break; }
      if (s[0] == '!') { skip = true; }
      else if (slen > 5) {
        if (strToUpper(s.substr(0, 6)) == "THERMO") {
          if (getOneLine(ifs, s) == ERRE) { break; }
          skip = true;
        }
      }
    }
    if (!skip) {
      vs.clear(); vs.push_back(s);
      for (id = 0; id < 3; id++) {
        if (getOneLine(ifs, s) == ERRE) { break; }
        vs.push_back(s);
      }
      nameSpc = strTermBy(vs[0], ' ');
      if (nameSpc.length() > 18) { nameSpc = nameSpc.substr(0, 18); }
      sSpOut += (' ' + nameSpc);
      if (sSpOut.length() > 60) { ofs << sSpOut << endl; sSpOut = ""; }
      oft << thmt[nameSpc].fix1stLine(vs[0], asLower) << endl;
      for (id = 1; id < 4; id++)
        { oft << thmt[nameSpc].fixDataLine(vs[id], id + 1) << endl; }
    }
  }
  if (sSpOut.length() > 0) { ofs << sSpOut << endl; }
  ifs.close();
  return NORM;
}

