/* -------------------------------------------------------------------------
     "joback.cc" - transport property estimation according to Joback
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2011 by A. Miyoshi, University of Tokyo
                                       created: Jan.  25, 2011
                                   last edited: Jan.  25, 2011
------------------------------------------------------------------------- */

#include "joback.h"

// ===== Joback group contribution value library class =====================

// ----- jobackValues ----------------------------------------------------

jobackValues::jobackValues() { clear(); }               // --- constructor

void jobackValues::clear() {                                  // --- clear
  tfpk = 0.; tbk = 0.; tck = 0.; pck = 0; vck = 0.;
}

// ----- jobackGroupValues -----------------------------------------------

jobackGroupValues::jobackGroupValues() { clear(); }     // --- constructor

int jobackGroupValues::restoreFrom(ifstream &ifs) {     // --- restoreFrom
  string s, gnam;
  vector<string> vs;
  jobackValues jv;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "jobackGroupValues{") { return ERRE; }
  clear();
  do {
    if (getTokenFrom(ifs, gnam) == ERRE) { return ERRE; }
    if (gnam != "}") {
      if (getTokensFrom(ifs, 5, vs) == ERRE) { return ERRE; }
      jv.tfpk = atof(vs[0].c_str());
      jv.tbk = atof(vs[1].c_str());
      jv.tck = atof(vs[2].c_str());
      jv.pck = atof(vs[3].c_str());
      jv.vck = atof(vs[4].c_str());
      insert(make_pair(gnam, jv));
    }
  } while (gnam != "}");
  return NORM;
}

// ----- jobackAliases ---------------------------------------------------

jobackAliases::jobackAliases() { clear(); }             // --- constructor

int jobackAliases::restoreFrom(ifstream &ifs) {         // --- restoreFrom
  string s, gnam, anam;
  vector<string> vs;
  if (getTokenFrom(ifs, s) == ERRE) { return ERRE; }
  if (s != "jobackAliases{") { return ERRE; }
  clear();
  do {
    if (getTokenFrom(ifs, gnam) == ERRE) { return ERRE; }
    if (gnam != "}") {
      if (getTokensFrom(ifs, 2, vs) == ERRE) { return ERRE; }
      anam = vs[1];
      insert(make_pair(gnam, anam));
    }
  } while (gnam != "}");
  return NORM;
}

// ----- jobackGrpList ---------------------------------------------------

jobackGrpList::jobackGrpList() { clear(); }             // --- constructor

// ----- jobackLibrary ---------------------------------------------------

jobackLibrary::jobackLibrary() { clear(); }             // --- constructor

void jobackLibrary::clear() {                                 // --- clear
  values.clear(); aliases.clear();
}

                                                           // --- loadFrom
int jobackLibrary::loadFrom(string flname, libPath &libp) {
  jobackGroupValues::iterator p;
  string gnam;
  ifstream ifs(libp.fullPath(flname).c_str());
  if (!ifs) { return ERRE; }
  if (values.restoreFrom(ifs) == ERRE) { ifs.close(); return ERRE; }
  cout << values.size() << " Joback groups loaded from [" << flname << "].\n";
  if (aliases.restoreFrom(ifs) == ERRE) { ifs.close(); return ERRE; }
  cout << aliases.size() << " group aliases loaded from [" << flname << "].\n";
  ifs.close();
  // - add group names to aliases
  for (p = values.begin(); p != values.end(); p++) {
    gnam = p->first;
    aliases.insert(make_pair(gnam, gnam));
  }
  return NORM;
}

                                                          // --- findGroup
int jobackLibrary::findGroup(string gnam, jobackValues &jv) {
  jobackAliases::iterator pa;
  jobackGroupValues::iterator pg;
  if ((pa = aliases.find(gnam)) == aliases.end()) { return ERRE; }
  if ((pg = values.find(pa->second)) == values.end()) { return ERRE; }
  jv = pg->second;
  return NORM;
}

                                                           // --- estimate
int jobackLibrary::estimate(jobackGrpList &jl, double &eps, double &sig) {
  jobackGrpList::iterator p;
  jobackValues jv;
  string gnam;
  int numg;
  double sTbk = 0., sTck = 0., sVck = 0.;
  double Tb, Tc, Vc;
  for (p = jl.begin(); p != jl.end(); p++) {
    gnam = p->first; numg = p->second;
    if (findGroup(gnam, jv) == ERRE)
      { cout << "unknown group [" << gnam << "].\n"; return ERRE; }
    sTbk += numg * jv.tbk; sTck += numg * jv.tck; sVck += numg * jv.vck;
  }
  Tb = 198. + sTbk;
  Tc = Tb / (0.584 + 0.965 * sTck - sTck * sTck);
  Vc = 17.5 + sVck;
  eps = Tc / 1.2593;
  sig = 0.809 * pow(Vc, 1./3.);
  return NORM;
}

