/*--------------------------------------------------------------------------
     "molecule.h" - molecular identification and property classes
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2010 by A. Miyoshi, University of Tokyo
                                       created: Aug.   3, 2002
                                        edited: Aug.  20, 2004
                                        edited: Feb.  21, 2010
                 - added classes molSMILES, exSMLSmol and related
                                   last edited: Jan.   7, 2011
--------------------------------------------------------------------------*/

#ifndef _MOLECULE_H_
#define _MOLECULE_H_

#include "element.h"
using namespace std;

#define FL_GRPMAI "grp_main.dat"        // main group library
#define FL_GRPCYC "grp_cyc.dat"         // cyclic group library
#define FL_GRPRAD "grp_rad.dat"         // radical group library
#define FL_ALKMCR "alk_macr.dat"        // alkane input macros
#define FL_TYPMOL "typ_mol.dat"         // molecular type id's

#define ALLALP    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

#define MAXN_ATOMS 150  // maximum number of (heavy) atoms in a molecule
 //   Now this constant is only used in bondInfo::getIdx(),
 //   and not used to declare arrays.
#define MAXN_BONDS 4    // maximum number of bonds from an atom

// ===== alkaneChain and related classes ===================================

class alkaneSubChain {
 public:                                         // --- data members ---
  int levSubst;         // Substitution level
  int parentChIdx;      // Index of parent chain
  int posSubInPCh;      // Subsitution position in the parent chain
  int lenChain;         // Length of the chain
 public:                                     // --- member functions ---
  alkaneSubChain();
  alkaneSubChain(int lS, int pCI, int pS, int lC);
  void clear();
};

class alkaneChain {
 public:                                         // --- data members ---
  string alkaneName;
  int isCyclic;
  int numChains;
  vector<alkaneSubChain> chain;
 protected:
  macroMap mcr;
 public:                                     // --- member functions ---
  alkaneChain();
  void clear();
  void addChain(alkaneSubChain sc);
  int setFrom(string &strInp);
  void print(ostream *os);
 private:
  int extrHead(string &s);
  int parseStr(int lev, int pIx, string &s);
};                                                    // --- globals ---
ostream &operator<<(ostream &os, alkaneChain &ac);

// ===== molecByChains and related classes =================================
//    chain structure of a molecule to find a unique representation

// ----- description of a chain - component of 'molecByChains' -------------

class atomChain {
 public:                                         // --- data members ---
  int isCyclic;
  int isSubChOfCyc;
  int numAtoms;
  vector<int> atomIndex;
  string sortKey;                  // "nnnNNAsB"
  string viewStr;
 public:                                     // --- member functions ---
  atomChain();
  atomChain(int ic, int isc, int n, vector<int> &ai, class molecInfo &m);
  int identLen(atomChain &ref);
  void print(ostream *os);
  int storeTo(ofstream &ofs);
  int restoreFrom(ifstream &ifs, class molecInfo &m);
  int isAromatic();
  int isIdxInChain(int idx);
 private:
  void set(int ic, int isc, int n, vector<int> &ai, class molecInfo &m);
};
                                                      // --- globals ---
ostream &operator<<(ostream &os, atomChain &ac);

// ----- tree-structuresd sub-chain description ----------------------------

class treeInfoChains {
 public:                                         // --- data members ---
  int numILsets;
  vector<int> numChsInILs;
  vector<int> identLOfILs;
  vector<int> numSubsInILs;
  vector<vector <int> > numChsInSub;
 public:                                     // --- member functions ---
  void addILset(int nC, int iL, int nS);
};

// ----- chain-structured representation of a molecule ---------------------

class molecByChains {
 public:                                         // --- data members ---
  int numChains;
  vector<atomChain> chain;
 public:                                     // --- member functions ---
  molecByChains();
  molecByChains(int start, int num, molecByChains &mc);
  void clear();
  void addChain(atomChain ac);
  void beUnique();
  void sortTree();
  void getRenumTbl(int n, vector<int> &sortTbl, vector<int> &trnsTbl);
  void renumAtoms(vector<int> &trnsTbl);
  void print(ostream *os);
  void getStrucView(vector<string> &vs);
  void addChainView(vector<string> &vs, int clrLen);
  void swapChains(int a, int b);
  int getCommonLen(int iCa, int iCb);
  int getChainOrigin(int iTch, int &pos);
  int operator>(molecByChains &rm);
  int operator<(molecByChains &rm);
  int operator==(molecByChains &rm);
  int operator!=(molecByChains &rm);
  int storeTo(ofstream &ofs);
  int restoreFrom(ifstream &ifs, class molecInfo &m);
  void copyPartial(int start, int num, molecByChains &mc);
 private:
  void replace(int start, int num, molecByChains &mc);
  int sortByKey();
  void getTreeInfo(treeInfoChains &ti);
};
                                                      // --- globals ---
ostream &operator<<(ostream &os, molecByChains &mc);


// ===== group additivity classes ==========================================
//     molecular representation in group additivity

// ----- group additivity information on neighborhood ----------------------

class groupNeighbor {
 public:                                         // --- data members ---
  int bondOrder;
  int toAtomicNbr;
  int UnsBnd2nd;        // 2/3 .. double/triple bonded to second neighbor
  int UB2ndAtomicNbr;   // Atomic num. of 2nd neighbor (for UnsBnd2nd > 0)
 public:                                     // --- member functions ---
  groupNeighbor();
  void clear();
  int operator<(groupNeighbor &robj);
  string getIdxStr();
  int restoreFrom(ifstream &ifs);
};

// ----- group additivity information on atom ------------------------------

class groupAtom {
 public:                                         // --- data members ---
  int atomicNumber;
  int atomAttrib;
  int numNeighbors;
  groupNeighbor neigh[MAXN_BONDS];
 public:                                     // --- member functions ---
  groupAtom();
  void clear();
  string getIdxStr();
  void sortNeighbors();
  int restoreFrom(ifstream &ifs);
  int isRadical();
  void beNonRadical();
};

// ----- map of G.A. info on atom ------------------------------------------

class groupAtomCFunc {                       // --- compare function ---
 public:
  int operator()(groupAtom lobj, groupAtom robj) const;
};

class groupMap : public map<groupAtom, string, groupAtomCFunc> {
 public:                                     // --- member functions ---
  groupMap();
  int loadFrom(string flname, libPath &libp);
  string findName(groupAtom &ga);
};

// ----- atom description in cyclic group ----------------------------------

class groupCycAtom {
 public:                                         // --- data members ---
  int atomicNumber;
  int atomAttrib;
  int bondOrdToNext;
 public:                                     // --- member functions ---
  groupCycAtom();
  void clear();
  string getIdxStrAtom();
  string getIdxStrBond();
  int restoreFrom(ifstream &ifs);
};

// ----- cyclic group ------------------------------------------------------

class groupCyclic {
 public:                                         // --- data members ---
  int numAtoms;
  vector<groupCycAtom> atom;
 public:                                     // --- member functions ---
  groupCyclic();
  void clear();
  string getIdxStr();
  void beUnique();
  int restoreFrom(ifstream &ifs);
};

// ----- map of cyclic groups ----------------------------------------------

class groupCyclicCFunc {                     // --- compare function ---
 public:
  int operator()(groupCyclic lobj, groupCyclic robj) const;
};

class groupCycMap : public map<groupCyclic, string, groupCyclicCFunc> {
 public:                                     // --- member functions ---
  groupCycMap();
  int loadFrom(string flname, libPath &libp);
  string findName(groupCyclic &gc);
};

// ----- all maps of groups ------------------------------------------------

class groupLib {
 public:                                         // --- data members ---
  groupMap main;
  groupCycMap cyc;
  groupMap rad;
 public:                                     // --- member functions ---
  groupLib();
  void clear();
  int loadFrom(string flnm, string flnc, string flnr, libPath &libp);
};

// ----- molecular representation by groups --------------------------------

class molecByGroups {
 public:                                         // --- data members ---
  string cycGrpName;
  string radGrpName;
  map<string, int> atmGrps;
 public:                                     // --- member functions ---
  molecByGroups();
  void clear();
  void print(ostream &os);
};
                                                       // --- globals ---
ostream &operator<<(ostream &os, molecByGroups &mg);

// ===== 'molecule' and related classes ====================================
//     representation of a molecule

// ----- chemical bond information by pair of atoms ------------------------

class chemBond {
 public:                                         // --- data members ---
  int order;
  int idxFirst;
  int idxSecond;
 public:                                     // --- member functions ---
  chemBond();
  chemBond(int ord, int idx1, int idx2);
  void setFrom(int ord, int idx1, int idx2);
  int operator==(chemBond &robj);
};

// ----- list of chemical bonds --------------------------------------------

class listBonds : public vector<chemBond> {
 public:                                     // --- member functions ---
  int addBond(chemBond bnd);
  int findBond(chemBond &bnd);
};

// ----- bonding information of an atom ------------------------------------

class bondInfo {
 public:                                         // --- data members ---
  int order;
  int toAtomicNbr;
  int toAtomIndex;
  bool isCommutable;
 public:                                     // --- member functions ---
  bondInfo();
  void clear();
  int operator>(bondInfo &robj);
  int operator<(bondInfo &robj);
  int getIdx();
  int storeTo(ofstream &ofs);
  int restoreFrom(ifstream &ifs);
  ostream &print(ostream &os);
};                                                    // --- globals ---
ostream &operator<<(ostream &os, bondInfo &bi);

// ----- atomic informations -----------------------------------------------

class atomInfo {
 public:                                         // --- data members ---
  int atomicNumber;
  int atomAttrib;
  int numBonds;
  bondInfo bond[MAXN_BONDS];
 public:                                     // --- member functions ---
  atomInfo();
  void clear();
  int operator>(atomInfo &robj);
  int operator<(atomInfo &robj);
  string getIdxStr();
  void sortBonds();
  int storeTo(ofstream &ofs);
  int restoreFrom(ifstream &ifs);
  vector<int> getBondToIdcs();
  bondInfo *bondTo(int idxAtom);
  int idxBondTo(int idxAtom);
  void removeBondTo(int idxAtom);
  void removeBond(int idxBond);
  int getNumHatoms();
  ostream &print(ostream &os);
};                                                    // --- globals ---
ostream &operator<<(ostream &os, atomInfo &ai);

// ----- molecular informations --------------------------------------------

class molecInfo {
 public:                                         // --- data members ---
  int numAtoms;
  vector<atomInfo> atom;
 public:                                     // --- member functions ---
  molecInfo();
  void clear();
  int operator==(molecInfo &rm);
  int operator!=(molecInfo &rm);
  int operator<(molecInfo &rm);
  int operator>(molecInfo &rm);
  void beUnique(molecByChains &unCh, int &isCyclic, int &nEqCh,
   int &isAroma);
  void renumAtoms(vector<int> &sortTbl, vector<int> &trnsTbl);
  int checkParam();
  void getAtomComp(atmcComp &ac);
  void getAtomCompSorted(atmcCompSorted &acs);
  string atmCompToFormula(atmcComp &ac);
  string getChemFormula();
  int getNumAtoms(int atmcNbr);
  void print(ostream *os);
  int storeTo(ofstream &ofs);
  int restoreFrom(ifstream &ifs);
  void beAromaOther(atomChain &ac);
  int getResonStr(vector<molecInfo> &rs);
  void beResonAvr(int numReson, vector<molecInfo> &rs);
  int getNumRadCent();
  int getIdxRadCent();
  void getIdcsRadCents(vector<int> &irc);
  string getStrRSE();
  void getGroupAroma(int idxAtom, atomChain &arc, groupAtom &ga);
  void getGroupNonAroma(int idxAtom, groupAtom &ga);
  void getDissocGroup(int idx, int cIdx, groupAtom &ga);
  int isRingPlanar(atomChain &ring);
  int getSymIsomCyclic(atomChain &ring, int &numIsom, int &numIntRot);
  int getSymIsomNonCyc(atomChain &mc, int neqc, int &numIsom,
   int &numIntRot);
  int getDistance(int id1, int id2);
  int getDstRoute(int id1, int id2, vector<int> &rt);
  int isPeroxyRad();
  int getIdxAtomMatch(int atmcNbr);
  int getIdxAtomMatch(int atmcNbr, int atmAttr);
  int getIdxAtomMatch(int atmcNbr, int toAtNbr, int bndOrd);
  int getIdxAtomMatch(int atmcNbr, int atmAttr, int toAtNbr, int bndOrd);
  int getIdxAtomMatch(int atmcNbr, int toAtNbr, int bndOrd, int to3AtNbr, int bnd3Ord);
  int getIdxNextAtom(int idxAtom, int atmcNbr);
  int getIdxNextAtom(int idxAtom, int atmcNbr, int bndOrd);
  int getIdxNextAtom(int idxAtom, int atmcNbr, int nxBndOrd, int nxAtmcNbr);
  int getIdxAtomTermMatch(int atmcNbr, int toAtNbr, int bndOrd);
  int getIdxAtomTermMatch(int atmcNbr, int atmAttr, int toAtNbr, int bndOrd);
  int getIdxAtomTermMatch(int atmcNbr, int bndOrd, int toAtNbr, int toBdOrd, int nxAtNbr);
  int getIdxAtomTermMatchNumH(int atmcNbr, int toAtNbr, int bndOrd, int toNumH);
  atomInfo *getAtomMatch(int atmcNbr);
  vector<int> getBondMatch(int an1, int an2, int bo);
  vector<vector<int> > getAllBondsMatch(int an1, int an2, int bo);
  int getAlleneCent(int &ia1, int &ia2);
  int getBondOrder(int ia1, int ia2);
  int countNextButOneC(int id);
  int numNeighQuatC(int id);
  int isQuatC(int id);
  int isQuatCincdO(int id);
  int isTertC(int id);
  int isTertCincdO(int id);
  void makeNewBond(int ia1, int ia2);
  void removeBond(int ia1, int ia2);
  void incBondOrder(int ia1, int ia2);
  void decBondOrder(int ia1, int ia2);
  void removeHatom(int idxAtom);
 private:
  void traceDistance(int cIdx, int idx, int &fIdx, int dst,
   set<int> &dstSet, set<int> passed);
  void traceDstRoute(int cIdx, int idx, int &fIdx, int dst,
   map<int, vector<int> > &dstMap, vector<int> route, set<int> passed);
  int compareWith(molecInfo &mi);
  void getRingSubList(atomChain &ring, vector<vector<int> > &sl,
   vector<int> &bor, vector<vector<int> > &isa);
  void getRotatedRepr(int nA, string repr, vector<string> *vs);
  void followBondsSub(int idx, int idxTo, listBonds &lb);
  int getIntRotSym(int cntPiv, int Piv, int &farLin);
  int getDBnRotSym(int cntPiv, int Piv);
  int countAllIntRotSym(int &intRotSym);
  string getListPlanarCent(atomChain &ring);
  int getPattMatchPlanCnt(string &sPlC, vector<string> *patt);
  int isTdSymmetric(atomChain &mc, int neqc, int &numIntRot,
   int &intRotSym);
  int isC3D3Symmetric(atomChain &mc, int neqc, int &numIntRot,
   int &intRotSym);
  int isC2D2Symmetric(atomChain &mc, int neqc, int &numIntRot,
   int &intRotSym);
  int isDBC2Symmetric(int &numIntRot, int &intRotSym);
  int isChiral();
  int isChiralCent(int idx);
  int isAlleneC2Sym(int idx1, int idx2);
  void sortAtoms();
  void swapRefs(int ir, int jr);
  int getTerminal();
  int followChain(int lev, int idx, molecByChains &molc);
  void listCyclics(int &numCases, atomChain &refc,
   vector<molecByChains> &molc);
  void followCycChain(int lev, int idxs, int idxe, int cidx,
    molecByChains &molc);
};
                                                      // --- globals ---
ostream &operator<<(ostream &os, molecInfo &m);

// ----- SMILES import class -----------------------------------------------

class atomSMILES : public atomInfo {
 public:                                         // --- data members ---
  int numRadcElc;       // number of radical electron
  int charge;
  int numDeloElc;       // number of delocalized electrons
  bool bracket;
  int numHexp;
  set<int> ringIdcs;
 public:                                     // --- member functions ---
  atomSMILES();
  void clear();
  void corrDeloc();
  int sumOrderBonds();
  int numBondingElec();
  int numHatoms();
  ostream &print(ostream &os);
};                                                    // --- globals ---
ostream &operator<<(ostream &os, atomSMILES &as);

class molSMILES {
 public:                                         // --- data members ---
  int numAtoms;
  int numRings;
  vector<atomSMILES> atoms;
 protected:
  string chBonds;
  string chOrgSb1st;
  set<string> str2cOrgSb;
  string sPars;
  int cpPars;
 public:                                     // --- member functions ---
  molSMILES();
  int setMol(string s, molecInfo &mi);
  bool isCharged();
 private:
  int parseChain(int pend, int nodBranch);
  int findEndPrnt(int pstart, int pend);
  void findBond(int pend, bondInfo &bnd);
  int findAtom(int pend, atomSMILES &atm);
  int parseBracketAtom(int pend, atomSMILES &atm);
  int resolvRingClosure();
  int resolvBrcktExp();
  void resolvDeloElc();
 public:
  ostream &print(ostream &os);
};
ostream &operator<<(ostream &os, molSMILES &ms);

// ----- SMILES export classes ---------------------------------------------

class exSMLSatom {
 public:                                         // --- data members ---
  int atomicNumber;
  int atomAttrib;
  int bondOrder;
  int numHatoms;
  int cycIndex;
  vector<int> idBranch;
 public:                                     // --- member functions ---
  exSMLSatom();
  void clear();
};

class exSMLSchain : public vector<exSMLSatom> {
 public:                                         // --- data members ---
  int branchedFrom;
  int branchPos;
 public:                                     // --- member functions ---
  exSMLSchain();
};

class exSMLSmol : public vector<exSMLSchain> {
 public:                                         // --- data members ---
  string strSMILES;
 protected:
  string chBonds;
  set<int> anOrgSb;
 public:                                     // --- member functions ---
  exSMLSmol();
  int setFrom(molecInfo &mi, molecByChains &mc);
  string getSMILES();
 protected:
  string chainSMLS(int cch);
  string Hstr(int nH);
};

// ----- typeMolec class ---------------------------------------------------

class typeMolec {
 public:                                         // --- data members ---
  map<long, string> mCodStr;
  map<string, long> mStrCod;
 public:                                     // --- member functions ---
  typeMolec();
  void clear();
  int loadFrom(string flname, libPath &libp);
  string typName(long tcod);
  string typNameExt(long tcod);
  long typCode(string tnam);
};

// ----- molecule ----------------------------------------------------------

class molecule {
 public:                                         // --- data members ---
  string molName;
  molecInfo molec;
  molecInfo resavr;
  molecByChains uniqueRepr;
  int isUnique;
  int isCyclic;
  int numEquivChain;
  int isAromatic;
  int numResStruct;
  int isRxnDone;
 public:                                     // --- member functions ---
  molecule();
  int setFrom(alkaneChain &alc);
  void setFrom(molecInfo &mi, string nam);
  int setFromPsmiles(string psm, groupLib &gl, typeMolec &tym);
  int setFromSMILES(string sm, string nam);
  string getSMILES();
  void clear();
  void print(ostream *os);
  void beUnique();
  void getUniqueView(vector<string> &vs);
  int storeTo(ofstream &ofs);
  void storeLogTo(ostream &os, groupLib &gl, typeMolec tm);
  int restoreFrom(ifstream &ifs);
  void getGroupAtom(int idxAtom, groupAtom &ga);
  void getGroupCyclic(groupCyclic &gc);
  void getMolecByGroups(molecByGroups &mg, groupLib &grp);
  int getIdxAtomGrpMatch(string gs, groupLib &grp);
  int getNumSymIsom(int &numIsom, int &numIntRot);
  string getThermInput(groupLib &grp);
  long getType(groupLib &gl);
  string getTypeFormula(groupLib &gl, typeMolec &tym);
  void beRadicalAt(int idxAtom);
  void beNonRadical();
  int isAtomInRing(int idxAtom);
  int getDstInRing(int id1, int id2, int &dstInRing);
  int getDstInRingRoute(int id1, int id2, int &dstInRing,
   vector<int> &rt);
  int nearestCH2noO(int idx);
  void moveOOH(int ixfrom, int ixto, int ixo);
  int isAldehyde();
  int isEnol(int &idCdOH, int &idCd, int &idOH);
  int isEnol();
  int isVinyl();
  int isAllyl();
  int isVinoxy();
  int isAcyl();
  int isBetaHOAlkyl();
};
                                                      // --- globals ---
ostream &operator<<(ostream &os, molecule &m);

#endif

