/*--------------------------------------------------------------------------
     "thermgen.c" - thermochemical data batch generator
        This is a part of KUCRS:
           Knowledge-basing Utilities for Complex Reaction Systems
                        (c) 2002-2010 by A. Miyoshi, University of Tokyo
                                       created: Oct.   3, 2001
                                        edited: Aug.  29, 2003
                                        edited: Dec.  10, 2003
                                        edited: Apr.  14, 2005
                                   last edited: Mar.   9, 2010
----------------------------------------------------------------------------
        Brief summary of changes made
         Revision 0.36 beta:
            - remove date string from thermodata
         Revision 0.20 -> 0.30:
            - extensively modified to ba a part of rxnKnob *
           1) program name changed from "genTherm" to "thermgen"
--------------------------------------------------------------------------*/

#ifndef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#endif

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#define ENVVARNAM "KUCRS_LIB"   /* environment variable name */

#define MXLenFN 256             /* maximum length of file names */
#define MXNfGAV 10              /* maximum number of GAV files */
#define MXNElem 20              /* maximum number of elements */
#define MXNGrp  800             /* maximum number of groups */
#define MXLenGN 16              /* maximum length of group name */
#define MXLenMN 19              /* maximum length of molecular name */
#define MXLenM1 11              /* max length of mol name (1st part) */
#define MXLenM2 9               /* max length of mol name (2nd part) */
#define MXLen1L 512             /* maximum length of one line */
#define MXLenHD 71              /* maximum length of header string */
#define NORM    0               /* normal end */
#define ERRE    (-1)            /* error end */
#define EEOF    1               /* normal EOF end */
#define TRUE    (-1)
#define FALSE   0
#define CONFGFN "thermcfg.dat"  /* configulation file name */
#define DELHF_H 52.10           /* heat of formation (298 K) of H */
#define CAL2J   4.184           /* calorie to Joule conversion */
#define RGAS    1.98722         /* gas constant in kcal mol-1 */

int configGAV(void);
void procGrp(int idx, char *buf);
void alltoUpper(char *buf);
double cut7val(int pos, char *buf);
int procHeader(FILE *inf, FILE *outf);
int proc1mol(FILE *inf, FILE *outf);
int getElemIdx(char *sym);
int getGrpIdx(char *sym);
void sepNameMol(char *nam, char *n1, char *n2);
int getALine(FILE *inf);

char namElem[MXNElem][3], namGrp[MXNGrp][MXLenGN];
char buf[MXLen1L], cbuf[MXLen1L];
int numElem, numGrp;
double grpH[MXNGrp], grpS[MXNGrp], grpCp[MXNGrp][7];

/*===== main =========================================================*/
int main(int argc, char *argv[]) {
 char inFile[MXLenFN], outFile[MXLenFN];
 int retProc;
 FILE *inf, *outf;
  if (argc != 3)
    { printf(" Usage: thermgen infile outfile\n"); return 1; }
  strcpy(inFile, argv[1]); strcpy(outFile, argv[2]);
  if (configGAV() == ERRE) return 1;
  if ((inf = fopen(inFile, "rt")) == NULL)
    { printf(" Failed to open %s!\n", inFile); return 1; }
  if ((outf = fopen(outFile, "wt")) == NULL) {
    printf(" Failed to open %s!\n", outFile);
    fclose(inf); return 1;
  }
  if (procHeader(inf, outf) == ERRE) {
    printf(" File I/O error!\n");
    fclose(inf); fclose(outf); return 1;
  }
  while ((retProc = proc1mol(inf, outf)) == NORM) ;
  fclose(outf);
  fclose(inf);
  if (retProc == ERRE) { return 1; }
  return 0;
}

/*===== GAV configulation ============================================*/
int configGAV(void) {
char libPath[MXLenFN], *plp, dum[20], fullfln[MXLenFN];
char flnGAV[MXNfGAV][MXLenFN];
int lenp, numGAVf, iGAVf, iElem, nGrp, iGrp, idxGrp;
FILE *conf, *gavf;
  if ((plp = getenv(ENVVARNAM)) == NULL) { strcpy(libPath, ""); }
  else {
    strcpy(libPath, plp);
    if ((lenp = strlen(libPath)) > 0) {
      if (libPath[lenp - 1] != '/') {
        libPath[lenp] = '/'; libPath[lenp + 1] = '\0';
      }
    }
  }
  strcpy(fullfln, libPath); strcat(fullfln, CONFGFN);
  printf(" --- reading a file : %s ", fullfln);
  if ((conf = fopen(fullfln, "rt")) == NULL)
    { printf(" Failed to open %s!\n", fullfln); return ERRE; }
  fscanf(conf, "%s %d", dum, &numGAVf);
  for (iGAVf = 0; iGAVf < numGAVf; ++iGAVf)
    { fscanf(conf, "%s", flnGAV[iGAVf]); }
  fscanf(conf, "%s %d", dum, &numElem);
  for (iElem = 0; iElem < numElem; ++iElem)
    { fscanf(conf, "%s", namElem[iElem]); }
  fclose(conf);
  printf(" - %d GAV-files and %d Elements\n", numGAVf, numElem);
  idxGrp = 0;
  for (iGAVf = 0; iGAVf < numGAVf; ++iGAVf) {
    strcpy(fullfln, libPath); strcat(fullfln, flnGAV[iGAVf]);
    if ((gavf = fopen(fullfln, "rt")) == NULL)
      { printf(" Failed to open %s!\n", fullfln); return ERRE; }
    printf(" --- reading a file : %s ", fullfln);
    getALine(gavf);
    do {
      getALine(gavf);
      strcpy(cbuf, strtok(buf, " ,"));
    } while (strchr("0123456789", cbuf[0]) == NULL);
    nGrp = atoi(cbuf);
    for (iGrp = 0; iGrp < nGrp; ++iGrp) {
      getALine(gavf);
      procGrp(idxGrp, buf);
      ++idxGrp;
    }
    fclose(gavf);
    printf(" - %d Groups\n", nGrp);
  }
  numGrp = idxGrp;
  printf(" --- total %d Groups\n", numGrp);
  return NORM;
}

/*----- process one group ----------------------------------*/
void procGrp(int idx, char *buf) {
int iCp;
  strcpy(cbuf, buf);
  strcpy(namGrp[idx], strtok(cbuf, " ,"));
  alltoUpper(namGrp[idx]);
  grpH[idx] = cut7val(14, buf);
  grpS[idx] = cut7val(21, buf);
  for (iCp = 0; iCp < 7; ++iCp) {
    grpCp[idx][iCp] = cut7val(28 + iCp * 7, buf);
  }
}

/*----- convert to all upper case --------------------------*/
void alltoUpper(char *buf) {
size_t lenbuf, pbuf;
  lenbuf = strlen(buf);
  for (pbuf = 0; pbuf < lenbuf; ++pbuf) {
    if (islower(buf[pbuf])) buf[pbuf] = toupper(buf[pbuf]);
  }
}

/*----- cut out seven chars and evaluate -------------------*/
double cut7val(int pos, char *buf) {
char s[8];
int i, ps, buflen, p;
double retv;
  buflen = strlen(buf);
  for (i = 0, ps = 0; i < 7; ++i) {
    p = pos + i;
    if (p >= buflen) break;
    if (strchr("0123456789.+-Ee", buf[p]) != NULL) s[ps++] = buf[p];
  }
  s[ps] = '\0';
  if (strlen(s) == 0) retv = -1001.; else retv = atof(s);
  return retv;
}

/*===== process header ===============================================*/
int procHeader(FILE *inf, FILE *outf) {
char headstr[MXLenHD], frmlen[5], frms[10];
int lenh;
  printf(" --- processing header\n");
  if (fgets(headstr, MXLenHD, inf) == NULL) return ERRE;
  lenh = strlen(headstr);
  if (lenh > 0)
    { if (headstr[lenh - 1] == '\n') headstr[lenh - 1] = '\0'; }
  if (fprintf(outf, "  UNITS:KJ\n") < 0) return ERRE;
  sprintf(frmlen, "%d", MXLenHD - 1);
  strcpy(frms, " %-"); strcat(frms, frmlen); strcat(frms, "s\n");
  fprintf(outf, frms, headstr);
  fprintf(outf, " SPECIES       Hf       S    Cp 300     400     500     ");
  fprintf(outf, "600     800     1000     1500     DATE        ELEMENTS\n");
  return NORM;
}

/*===== process one molecule =========================================*/
int proc1mol(FILE *inf, FILE *outf) {
char namMol[MXLenMN], symElem[3], symGrp[MXLenGN];
char namM1[MXLenM1], namM2[MXLenM2];
char *nam;
int IsRadical, nkElem, elemIdx[4], elemNum[4], i, iCp;
int nkGrp, grpIdx[20], grpNum[20], grpRadIdx, numRot, numSym;
double molH, molS, molCp[7];
  /*--- read inputs ------------------------------*/
  do {
    if (getALine(inf) == ERRE) return EEOF;
    strcpy(cbuf, buf);
    nam = strtok(cbuf, " ,\t");
    if (nam == NULL) namMol[0] = '\0';
      else strncpy(namMol, nam, MXLenMN - 1);
  } while ((namMol[0] == '\0') || (namMol[0] == '!'));
  printf(" - proccessing species %s ", namMol);
  alltoUpper(buf);
  strtok(buf, " ,\t");
  if (strcmp(strtok(NULL, " ,\t"), "R") == 0) IsRadical = TRUE;
   else IsRadical = FALSE;
  nkElem = 0;
  for (i = 0; i < 4; ++i) {
    strcpy(symElem, strtok(NULL, " ,\t"));
    if (strcmp(symElem, "$") == 0) break;
    elemIdx[nkElem] = getElemIdx(symElem);
    if (elemIdx[nkElem] < 0)
      { printf("- unknown element. - ABORT!\n"); return ERRE; }
    elemNum[nkElem] = atoi(strtok(NULL, " ,\t"));
    ++nkElem;
  }
  if (strcmp(symElem, "$") != 0) {
    strcpy(symElem, strtok(NULL, " ,\t"));
    if (strcmp(symElem, "$") != 0)
      { printf("- too many elements. - ABORT!\n"); return ERRE; }
  }
  nkGrp = 0;
  for (i = 0; i < 20; ++i) {
    strcpy(symGrp, strtok(NULL, " ,\t"));
    if (strcmp(symGrp, "$") == 0) break;
    grpIdx[nkGrp] = getGrpIdx(symGrp);
    if (grpIdx[nkGrp] < 0)
      { printf("- unknown group. - ABORT!\n"); return ERRE; }
    grpNum[nkGrp] = atoi(strtok(NULL, " ,\t"));
    ++nkGrp;
  }
  if (strcmp(symGrp, "$") != 0) {
    strcpy(symGrp, strtok(NULL, " ,\t"));
    if (strcmp(symGrp, "$") != 0)
      { printf("- too many groups. - ABORT!\n"); return ERRE; }
  }
  if (IsRadical) {
    strcpy(symGrp, strtok(NULL, " ,\t"));
    grpRadIdx = getGrpIdx(symGrp);
    if (grpRadIdx < 0)
      { printf("- unknown group. - ABORT!\n"); return ERRE; }
  }
  numRot = atoi(strtok(NULL, " ,\t"));
  numSym = atoi(strtok(NULL, " ,\t"));
  /*--- GA calculation ---------------------------*/
  molH = 0.; molS = 0.;
  for (iCp = 0; iCp < 7; ++iCp) molCp[iCp] = 0.;
  for (i = 0; i < nkGrp; ++i) {
    molH += (grpH[grpIdx[i]] * grpNum[i]);
    molS += (grpS[grpIdx[i]] * grpNum[i]);
    for (iCp = 0; iCp < 7; ++iCp) {
      molCp[iCp] += (grpCp[grpIdx[i]][iCp] * grpNum[i]);
    }
  }
  if (IsRadical) {
    molH += grpH[grpRadIdx] - DELHF_H;
    molS += grpS[grpRadIdx];
    for (iCp = 0; iCp < 7; ++iCp) {
      molCp[iCp] += grpCp[grpRadIdx][iCp];
    }
  }
  molS -= RGAS * log((double)numSym);
  molH *= CAL2J; molS *= CAL2J;
  for (iCp = 0; iCp < 7; ++iCp) molCp[iCp] *= CAL2J;
  /*--- output -----------------------------------*/
  sepNameMol(namMol, namM1, namM2);
  fprintf(outf, " %-10s%8.2f %8.2f ", namM1, molH, molS);
  for (iCp = 0; iCp < 7; ++iCp) {
    if (molCp[iCp] < -1000.) fprintf(outf, "        ");
      else fprintf(outf, " %7.2f", molCp[iCp]);
  }
  fprintf(outf, " %-8s", namM2);
  fprintf(outf, "thgen   ");
  for (i = 0; i < nkElem; ++i)
    fprintf(outf, " %-2s%3d", namElem[elemIdx[i]], elemNum[i]);
  for (i = nkElem; i < 4; ++i) fprintf(outf, "     0");
  fprintf(outf, " G%2d\n", numRot);
  printf("- done.\n");
  return NORM;
}

/*----- get element index ----------------------------------*/
int getElemIdx(char *sym) {
int iElem;
  for (iElem = 0; iElem < numElem; ++iElem) {
    if (strcmp(sym, namElem[iElem]) == 0) return iElem;
  }
  return (-1);
}

/*----- get group index ------------------------------------*/
int getGrpIdx(char *sym) {
int iGrp;
  for (iGrp = 0; iGrp < numGrp; ++iGrp) {
    if (strcmp(sym, namGrp[iGrp]) == 0) return iGrp;
  }
  return (-1);
}

/*----- separate molecular name into 10 + 8 chars ----------*/
void sepNameMol(char *nam, char *n1, char *n2) {
int lennam, lenpls, i;
  lennam = strlen(nam);
  if (lennam < 11) {
    strcpy(n1, nam); n2[0] = '\0';
  } else {
    strncpy(n1, nam, 10); n1[10] = '\0';
    lenpls = lennam - 10;
    for (i = 0; i < lenpls; ++i) n2[i] = nam[i + 10];
    n2[lenpls] = '\0';
  }
}

/*----- get a line ---------------------------------------------------*/
int getALine(FILE *inf) {
size_t lens;
  if (fgets(buf, MXLen1L, inf) == NULL) return ERRE;
  lens = strlen(buf);
  if (lens > 0) {
    if (buf[lens - 1] == '\n') buf[lens - 1] = '\0';
  }
  return NORM;
}

